"""
Provides fdsreader version information.
"""

# This file is auto-generated! Do not edit!
# Use `python -m incremental.update fdsreader` to change this file.

from incremental import Version

__version__ = Version("fdsreader", 0, 9, 15)
__all__ = ["__version__"]
